package files;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.remote.RemoteWebDriver;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author V. Mihashcenko
 * @version 1.1
 *
 * The class offers methods to get companies data of Linkedin website
 * and 2 other recources based on given links.
 * It also allowes to get a list of urls from the list page
 */

public class ScrapeLinkedinCompanyData
{
    private final RemoteWebDriver driver;
    private String userAgent = "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Safari/537.36 OPR/68.0.3618.129";
    private HashMap<String, String> crunchBaseCookie;
    private int maxFale = 10;

    public RemoteWebDriver getDriver() { return driver; }
    public String getUserAgent() { return userAgent; }
    public HashMap<String, String> getCrunchBaseCookie() { return crunchBaseCookie; }

    public void setUserAgent(String userAgent) { this.userAgent = userAgent; }
    public void setMaxFale(int maxFale) { this.maxFale = maxFale; }

    public void setCrunchBaseCookie(HashMap<String, String> crunchBaseCookie) { this.crunchBaseCookie = crunchBaseCookie; }

    /* ************************************************************************************ */

    /**
     * Iterates over the links from a file, each link should be in a new line 
     * @param txtWithLinks - text file with the list of links.
     * @param deleteSuccessURL - command to delete successfully scraped urls. File will be fully rewritten!
     * @return a number of successfully scraped links/urls.
     */
    public ArrayList<LinkedinData> scrapeFromFile(File txtWithLinks, boolean deleteSuccessURL) throws IOException
    {
        if(txtWithLinks == null)
            throw new NullPointerException("txtWithLinks = null");

        int scrapeFale = 0; // number of failed urls
        HashSet<String> url = new HashSet<>(); // list of urls
        ArrayList<LinkedinData> result = new ArrayList<>();

        try(BufferedReader buffR = new BufferedReader(new FileReader(txtWithLinks)))
        {
            String str;
            while ((str = buffR.readLine()) != null)
                url.add(str);
        }

        for(String str : url)
        {
            if(scrapeFale >= maxFale)
                break;

            LinkedinData linkedinData = scrapeData(str);
            if(linkedinData != null)
            {
                result.add(linkedinData);
                url.remove(str);
            }
            else
                scrapeFale++;
        }

        if(deleteSuccessURL)
        {
            try(BufferedWriter buffW = new BufferedWriter(new FileWriter(txtWithLinks)))
            {
                for(String s : url)
                    buffW.write(s + '\n');
            }
        }

        return result;
    }

    /* ************************************************************************************ */

    /**
     * Gets data from 
	 *  - linkedin.com
     *  - crunchbase.com (email and phone)
     *  - bing.com (langitude, latitude, code(index) and address)
     * JSoup is used along with Selenium
     */
    public LinkedinData scrapeData(String url) throws IOException
    {
        if(url == null)
            throw new NullPointerException("url = null");

        driver.navigate().to(url);
        System.out.println("#### Start look for data (" + url + ") ####");

        List<WebElement> webElements;
        Elements elements;

        Document html = Jsoup.parse(driver.getPageSource());

        LinkedinData result;

        /* Using JSoup ot parse data */

        elements = html.select("h1[title] span");
        if(elements.size() > 0)
            result = new LinkedinData(elements.get(0).text().trim());
        else
            return null;

        elements = html.select("dl dt:nth-child(3)");
        if(elements.size() > 0 && elements.get(0).text().trim().equals("Industry"))
        {
            elements = html.select("dl dd:nth-child(4)");
            if(elements.size() > 0)
                result.setIndustry(elements.get(0).text().trim());
            else
                return null;
        }
        else
            return null;

        elements = html.select("a[data-control-name=\"page_details_module_website_external_link\"]");
        if(elements.size() > 0)
            result.setLink(elements.get(0).attr("href").trim());

        elements = html.select("dd.org-about-company-module__company-size-definition-text");
        if(elements.size() > 0)
            result.setEmployeesRange(elements.get(0).text().trim());

        elements = html.select("dd.org-page-details__employees-on-linkedin-count");
        if(elements.size() > 0)
        {
            String employeesNum = elements.get(0).text();
            Pattern pattern = Pattern.compile("[0-9,]+(?= on LinkedIn)");
            Matcher matcher = pattern.matcher(employeesNum);
            if(matcher.find())
                result.setEmployeesNum(matcher.group().trim());
        }

        elements = html.select("dl dd:last-of-type");
        if(elements.size() > 0)
            result.setBussinessTypes(elements.get(0).text().trim());

        String linkToCrunchBase = null;
        elements = html.select("a[data-control-name=\"funding_crunchbase_company_see_more_link\"]");
        if(elements.size() > 0)
            linkToCrunchBase = elements.get(0).attr("href").trim();

        /* Search of link for the "bing" map */
        ArrayList<String> linkToMap = new ArrayList<>();
        ArrayList<String> linkToMapFromMethod;
        boolean findMainOffice = false;

        /* Variant 1 - links for the map are already there  */
        linkToMapFromMethod = lookForLinksToMap();
        if(linkToMapFromMethod != null)
        {
            if(!linkToMapFromMethod.get(0).equals(""))
            {
                linkToMap.add(0, linkToMapFromMethod.get(0));
                findMainOffice = true;
            }

            for(int k = 1; k < linkToMapFromMethod.size(); k++)
                linkToMap.add(linkToMapFromMethod.get(k));
        }
        else
        {
            /* Variant 2 - world map with the points/dots on it */
            webElements = driver.findElementsByCssSelector("g.highcharts-series-group > g:last-of-type path[role]");
            if(webElements.size() > 0)
            {
                List<WebElement> webElementsLoop;

                /* iterate over all the points on the map */
                for(int i = 0; i < webElements.size(); i++)
                {
                    try
                    {
                        webElements.get(i).click();
                    }
                    catch(Exception e){/* an exception will be triggered since some of used elements get disappeared */}

                    linkToMapFromMethod = lookForLinksToMap();
                    if(linkToMapFromMethod != null)
                    {
                        if(!linkToMapFromMethod.get(0).equals(""))
                        {
                            linkToMap.add(0, linkToMapFromMethod.get(0));
                            findMainOffice = true;
                        }

                        for(int k = 1; k < linkToMapFromMethod.size(); k++)
                            linkToMap.add(linkToMapFromMethod.get(k));
                    }

                    /* back to the global map */
                    webElementsLoop = driver.findElementsByCssSelector("button.org-locations-module__close-left-panel-button");
                    if(webElementsLoop.size() > 0)
                    {
                        webElementsLoop.get(0).click();
                        try {
                            /* map points don't have time to get displayed  */
                            Thread.sleep(1000);
                        }
                        catch(Exception e) {/* an exception will be triggered since some of used elements get disappeared */}
                        webElements = driver.findElementsByCssSelector("g.highcharts-series-group > g:last-of-type path[role]");
                    }
                    else
                        break;
                }
            }
        }

        /* actions at cruncbase.com. Search for an email and phone number. Use of JSoup */
        if(linkToCrunchBase != null)
            lookForNumberAndEmail(result);

        /* Search for address */
        if(linkToMap.size() > 0)
        {
            for(int i = 0; i < linkToMap.size(); i++)
            {
                LinkedinAddressData linkedinAddressData = scrapeAddress(linkToMap.get(i));
                if(linkedinAddressData != null)
                {
                    if(i == 0 && findMainOffice)
                        linkedinAddressData.setMainOffice(true);
                    result.getAddressData().add(linkedinAddressData);
                }
            }
        }

        return result;
    }

    /**
     * Search number and email from crunchbase.com
     */
    public void lookForNumberAndEmail(LinkedinData linkedinData) throws IOException
    {
        Document crunchBasePage = Jsoup.connect("https://www.crunchbase.com/organization/microsoft")
                .cookies(crunchBaseCookie)
                .userAgent(userAgent)
                .get();

        /* Search for email and phone number */
        Elements elements = crunchBasePage.select("span.field-value");
        boolean emailFind = false;
        boolean numberFind = false;
        if(elements.size() > 0)
        {
            for(Element elem : elements)
            {
                String text = elem.text().trim();
                if(!emailFind)
                {
                    if(text.matches("[a-zA-Z0-9.\\-_]+@[a-zA-Z]+[.]{1}[a-zA-Z]{2,4}"))
                    {
                        emailFind = true;
                        linkedinData.setEmail(text);
                    }
                }

                if(!numberFind)
                {
                    if(text.matches("[0-9 /.)(+\\-\\\\]{7,}"))
                    {
                        numberFind = true;
                        linkedinData.setNumber(text);
                    }
                }

                if(emailFind && numberFind)
                    break;
            }
        }
    }

    /**
     * Search for link to the map at "bing"
     * ! In the returned collection the first element (position 0) is a link to the main office 
     *   if main office is not found the first element contains an empty line ("").
     */
    public ArrayList<String> lookForLinksToMap()
    {
        List<WebElement> webElements;
        ArrayList<String> result = new ArrayList<>();
        boolean primaryFind = false;

        webElements = driver.findElementsByCssSelector("a[data-control-name=\"locations_module_directions_link\"]");
        if(webElements.size() > 0)
        {
            for(WebElement elem : webElements)
                result.add(elem.getAttribute("href"));

            /* Search of the main office */
            List<WebElement> primary = driver.findElementsByCssSelector("span[aria-label=\"This location is the primary office.\"] ~ a");
            if(primary.size() > 0)
            {
                primaryFind = true;
                String primaryLink = primary.get(0).getAttribute("href");
                result.remove(primaryLink);
                result.add(0, primaryLink);
            }
        }

        if(!primaryFind)
            result.add(0, "");

        if(result.size() == 1 && result.get(0).equals(""))
            return null;

        return result;
    }

    /**
     * @param url link to the bing map
     */
    public LinkedinAddressData scrapeAddress(String url)
    {
        if(url == null)
            throw new NullPointerException("url = null");

        driver.navigate().to(url);

        List<WebElement> webElements;
        LinkedinAddressData result;

        /* search of address, city, country and zip code */
        webElements = driver.findElementsByCssSelector("#maps_sb");
        if(webElements.size() > 0)
        {
            /* we take address from 2 sources and compares it using better records of the place.
             * basic: addressFromInput */
            String[] addressFromInput = webElements.get(0).getAttribute("value").split(",");
            String[] addressFromDiv = null;
            webElements = driver.findElementsByCssSelector("div.nameContainer");
            if(webElements.size() > 0)
            {
                addressFromDiv = webElements.get(0).getText().split(",");
                if(addressFromDiv.length == addressFromInput.length)
                {
                    for(int i = 0; i < addressFromInput.length; i++)
                    {
                        if(addressFromDiv[i].length() > addressFromInput[i].length())
                            addressFromInput[i] = addressFromDiv[i];
                    }
                }
            }
            result = new LinkedinAddressData(String.join(",", addressFromInput));

            /* get data from the acquired string */
            if(addressFromInput.length >= 3 && addressFromInput.length <= 5)
            {
                int size = addressFromInput.length;
                result.setCountry(addressFromInput[size - 1].trim());
                result.setCity(addressFromInput[size - 2].trim());

                Pattern pattern = Pattern.compile("[0-9\\-]+");
                Matcher matcher = pattern.matcher(addressFromInput[size - 2]);
                if(matcher.find())
                {
                    result.setCode(matcher.group());
                    result.setCity(addressFromInput[size - 2].replaceFirst(matcher.group(), "").trim());
                }
            }

            /* coordinates search */
            webElements = driver.findElementsByCssSelector("div.geochainModuleLatLong");
            String[] coord = null;
            if(webElements.size() > 0)
                coord = webElements.get(0).getText().split(",");
            if(coord != null && coord.length == 2)
            {
                result.setLatitude(coord[0].trim());
                result.setLongitude(coord[1].trim());
            }

            return result;
        }
        else
            return null;
    }

    /* ************************************************************************************ */

    /**
     * Get the links to the companies pages from linkedin.com/directory/companies...
     */
    public ArrayList<String> scrapeCompaniesURL(String url)
    {
        if(url == null)
            throw new NullPointerException("url = null");

        driver.navigate().to(url);

        List<WebElement> links = driver.findElementsByCssSelector("li.content a");
        ArrayList<String> result = new ArrayList<>(2000);
        for(WebElement elem : links)
        {
            String link = elem.getAttribute("href");

            /* remove from the link redundant part and make a company description link */
            if(link.matches(".+?\\?.+"))
                link = link.replaceAll("\\?.+", "/about");
            else
                link += "/about";

            result.add(link);
        }
        result.trimToSize();
        return result;
    }

    /* ************************************************************************************ */

    public ScrapeLinkedinCompanyData(RemoteWebDriver driver) { this.driver = driver; }
    public ScrapeLinkedinCompanyData(RemoteWebDriver driver, HashMap<String, String> crunchBaseCookie) {this.driver = driver; this.crunchBaseCookie = crunchBaseCookie;}
}
